from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.datasets.dataloader_factory import register_dataset
import numpy as np
from typing import Any
from nsquared.data_types import DataType

params = {
    "num_rows": (int, 100, "Number of rows in the dataset"),
    "num_cols": (int, 100, "Number of columns in the dataset"),
    "seed": (int, None, "Random seed for reproducibility"),
    "stddev_noise": (float, 1, "Standard deviation of noise"),
    "snr": (
        float,
        None,
        "Signal-to-noise ratio (defaults to None—if provided, overrides stddev_noise)",
    ),
    "mode": (str, "mcar", "Missing data mechanism ('mcar' or 'mnar')"),
    "miss_prob": (float, 0.5, "Probability of missing values (for MCAR)"),
    "mnar_deter": (
        bool,
        True,
        "Whether to use deterministic observation probabilities for MNAR",
    ),
    "latent_factor_combination_model": (
        str,
        "multiplicative",
        "Additive or multiplicative model for combining latent factors",
    ),
    "rho": (
        float,
        0.5,
        "Parameter for data generation in the nonlinear additive model. I.e. lambda in the Holder signal",
    ),
    "latent_factor_dimensionality": (int, 4, "Dimensionality of the latent factors"),
    "simulated_data_nonlin_transform": (
        str,
        "",
        "Nonlinear transformation to apply to the synthetic data, if any",
    ),
}


@register_dataset("synthetic_data", params)
class SyntheticDataLoader(NNDataLoader):
    """Data from the Heartsteps V1 study formatted into a matrix or tensor.
    To initialize with default settings, use: NNData.create("synthetic")
    """

    def __init__(
        self,
        download: bool = False,  # this is not needed for synthetic data but is kept for consistency
        save_dir: str = "./",
        agg: str = "mean",
        save_processed: bool = False,
        # above are superconstructor args
        # below are specific args
        num_rows: int = 100,
        num_cols: int = 100,
        mode: str = "mcar",
        rho: float = 0.5,
        seed: int | None = None,
        stddev_noise: float = 1,
        snr: float | None = None,
        miss_prob: float = 0.5,
        mnar_deter: bool = True,
        latent_factor_combination_model: str = "multiplicative",
        latent_factor_dimensionality: int = 4,
        simulated_data_nonlin_transform: str = "",
        **kwargs: Any,
    ):
        """Initializes the Synthetic data loader.

        Args:
        ----
            download: Whether to download the data locally. Default: False. If True, data is downloaded at save_dir
            save_dir: Directory to download the data to or where it already exists. Also the directory where the processed data will be. Default: "./" (current directory).
            agg: Aggregation method to use to create scalar dataset. Default: "mean".
            save_processed: Whether to save the processed data. Default: False.
            num_rows: Number of rows in the dataset. Default: 100.
            num_cols: Number of columns in the dataset. Default: 100.
            mode: Missing data mechanism ('mcar' or 'mnar'). Default: "mcar".
            rho: Parameter for data generation, or lambda in the Holder signal. Default: 0.5.
            seed: Random seed for reproducibility. Default: None.
            stddev_noise: Standard deviation of noise. Default: 1.
            snr: Signal-to-noise ratio. Default: None. If provided, overrides stddev_noise.
            miss_prob: Probability of missing values (for MCAR). Default: 0.5.
            mnar_deter: Whether to use deterministic observation probabilities for MNAR. Default: True.
            latent_factor_combination_model: Additive or multiplicative model for combining latent factors. Default: "multiplicative".
            latent_factor_dimensionality: Dimensionality of the latent factors. Default: 4.
            simulated_data_nonlin_transform: Nonlinear transformation to apply to the synthetic data, if any. Default: "", i.e. no transformation.
            kwargs: Additional keyword arguments.

        """
        super().__init__(
            download=download,
            save_dir=save_dir,
            agg=agg,
            save_processed=save_processed,
            **kwargs,
        )
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.mode = mode
        self.rho = rho
        self.stddev_noise = stddev_noise
        self.snr = snr  # if provided, this overrides stddev_noise
        self.miss_prob = miss_prob
        self.mnar_deter = mnar_deter
        self.latent_factor_combination_model = (
            latent_factor_combination_model  # "multiplicative" or "additive"
        )
        self.latent_factor_dimensionality = latent_factor_dimensionality
        self.simulated_data_nonlin_transform = simulated_data_nonlin_transform
        self.init_seed = seed
        if seed is not None:
            np.random.seed(
                seed=seed
            )  # instantiate random seed if provided but do it only once here

    def download_data(self) -> None:
        """Nothing to download for synthetic data"""
        print(
            "Warning: 'download' arg set to true, but there is no data to download when using synthetic data."
        )
        pass

    def _generate_simulated_data(self) -> None:
        """Generates the simulated data with no missing values"""
        # row and column latent factors:
        U = (
            np.random.uniform(size=(self.num_rows, self.latent_factor_dimensionality))
            - 0.5
        )  # U ~ Uniform(-0.5, 0.5), shape (num_rows, latent_factor_dimensionality)
        V = (
            np.random.uniform(size=(self.num_cols, self.latent_factor_dimensionality))
            - 0.5
        )  # V ~ Uniform(-0.5, 0.5), shape (num_cols, latent_factor_dimensionality)

        # Let us define a tensor Y constructed from U and V
        if self.latent_factor_combination_model == "additive":
            print("Additive model")
            # ... such that Y_i,j = |U_i + V_j|^rho * sign(U_i + V_j)
            # This Holder Continuous f(U,V), allows for a potentially non-linear relationship between U and V
            # vide: https://doi.org/10.48550/arXiv.2411.12965
            Y = np.abs(U[:, np.newaxis] + V) ** self.rho * np.sign(
                U[:, np.newaxis] + V
            )  # shape (latent_factor_dimensionality, latent_factor_dimensionality)
            Y = Y.sum(
                axis=2
            )  # collapse the latent factor dimensionality to get Y of shape (num_rows, num_cols)
        elif self.latent_factor_combination_model == "multiplicative":
            # ...such that Y_i,j = U_i * V_j
            Y = U @ V.T
        else:
            raise ValueError("Invalid latent_factor_combination_model arg provided.")
        # print("Y shape: ", Y.shape)
        self._transform_simulated_data(Y)
        data_true = Y.copy()  # i.e. "Theta"

        if (
            self.snr is not None
        ):  # if signal-to-noise ratio provided, override stddev_noise
            signal_var = np.mean(Y**2)
            stddev_noise = np.sqrt(signal_var / self.snr)
            self.stddev_noise = (
                stddev_noise  # update the instance variable for future reference
            )
        Y += self.stddev_noise * np.random.normal(size=(self.num_rows, self.num_cols))
        data_noisy = Y.copy()

        self.row_latent = U
        self.data_true = data_true
        self.data_noisy = data_noisy
        self.col_latent = V

    def _make_mcar(self) -> None:
        """Makes values missing completely at random (MCAR)"""
        # M ~ Bernoulli(self.miss_prob), shape (self.num_rows, self.num_cols)
        missing_mask = (
            np.random.binomial(1, self.miss_prob, size=(self.num_rows, self.num_cols))
            == 1
        )
        data_obs = self.data_noisy.copy()
        data_obs[missing_mask] = np.nan
        A = ~missing_mask  # A = NOT M, i.e. A_ij = 1 if Y_ij is observed, 0 if missing
        self.data_obs = data_obs
        self.availability_mask = A

    def _make_mnar(self) -> None:
        raise NotImplementedError("MNAR yet to be implemented")
        # TODO: Aashish/Caleb/Kyuesong/Tatha: come back to this later

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state of this object as a dictionary"""
        return_dict = {
            "data_observed": self.data_obs,
            "observed_entries": self.availability_mask,
            "full_data_true": self.data_true,
            "full_data_noisy": self.data_noisy,
            "row_latent": self.row_latent,
            "col_latent": self.col_latent,
        }
        if include_metadata:
            return_dict["generation_metadata"] = {
                "num_rows": self.num_rows,
                "num_cols": self.num_cols,
                "mode": self.mode,
                "stddev_noise": self.stddev_noise,
                "snr": self.snr,
                "miss_prob": self.miss_prob,
                "mnar_deter": self.mnar_deter,
                "latent_factor_combination_model": self.latent_factor_combination_model,
                "latent_factor_dimensionality": self.latent_factor_dimensionality,
                "simulated_data_nonlin_transform": self.simulated_data_nonlin_transform,
                "seed_at_generation": self.init_seed,
                "rho": self.rho,
            }
        return return_dict

    def run_simulation(self) -> None:
        """Runs the simulation and makes data missing as needed"""
        self._generate_simulated_data()
        if self.mode == "mcar":
            self._make_mcar()
        elif self.mode == "mnar":
            self._make_mnar()

    def process_data_scalar(
        self,
        agg: str = "mean",  # TODO Caleb, is it safe to remove these args?
    ) -> tuple[np.ndarray, np.ndarray]:
        """Forces a simulation run, and processes synthetic data as scalar form.

        Args:
        ----
            agg: Aggregation method to use. Default: "mean".

        Returns:
        -------
            tuple: Tuple containing:
                - np.ndarray: Data matrix with missing values as NaN
                - np.ndarray: Binary availability mask (1 for observed, 0 for missing)

        """
        self.run_simulation()
        state_dict = self.get_full_state_as_dict(include_metadata=False)
        observed_data = state_dict["data_observed"]
        availability_mask = state_dict["observed_entries"]
        return (observed_data, availability_mask)

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Method not yet implemented for synthetic data

        Args:
            data_type: Data type to process. Default: None.

        """
        # TODO Caleb, Kyuesong: is this even well-defined for synthetic data?
        raise NotImplementedError(
            "Distributional setting not yet implemented for synthetic data"
        )

    # HELPER FUNCTIONS
    def _expit(self, x: np.ndarray) -> np.ndarray:
        """Helper function to apply the logistic sigmoid function to an array"""
        return np.exp(x) / (1 + np.exp(x))

    def _transform_simulated_data(self, Y: np.ndarray) -> np.ndarray:
        """Apply a nonlinear transformation to the simulated data. Stored here since it's messy to have it in the main simulation."""
        non_lin = self.simulated_data_nonlin_transform
        if non_lin == "":
            return Y
        if non_lin == "expit":  # logistic sigmoid
            Y = self._expit(Y)
        elif non_lin == "tanh":
            Y = np.tanh(Y)
        elif non_lin == "sin":
            Y = np.sin(Y)
        elif non_lin == "cubic":
            Y = Y**3
        elif non_lin == "sinh":
            Y = np.sinh(Y)
        else:
            raise ValueError(
                "non_lin must be one of '', 'expit', 'tanh', 'sin', 'cubic', or 'sinh'."
            )
        return Y



"""Utility functions for the experiment scripts"""

from argparse import ArgumentParser
import logging
import numpy as np


def get_base_parser() -> ArgumentParser:
    """Get the base CLI parser for the heartsteps scripts"""
    parser = ArgumentParser()
    parser.add_argument(
        "--output_dir", "-od", type=str, default="out", help="Output directory"
    )
    parser.add_argument(
        "--estimation_method",
        "-em",
        type=str,
        default="row-row",
        choices=[
            "dr",
            "ts",
            "row-row",
            "col-col",
            "usvt",
            "softimpute",
            "auto",
            "aw",
            "usvt",
        ],
        help="Estimation method to use",
    )
    parser.add_argument(
        "--fit_method",
        "-fm",
        type=str,
        default="lbo",
        choices=["dr", "ts", "lbo", "usvt"],
        help="Fit method to use",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Force overwrite of existing results",
    )
    parser.add_argument(
        "--seed",
        "-s",
        type=int,
        default=42,
        help="Random seed",
    )
    parser.add_argument(
        "--log_level",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Log level",
    )
    parser.add_argument(
        "--allow_self_neighbor", action="store_true", help="Allow self neighbor"
    )
    parser.add_argument(
        "--raw_threshold",
        action="store_true",
        help="Use raw (not percentile-based) for distance threshold",
    )

    parser.add_argument(
        "--propensity",
        "-p",
        type=float,
        default=0.5,
        help="Propensity for the missing data",
    )
    return parser


def setup_logging(log_level: str) -> None:
    """Setup logging for the experiment scripts

    Args:
        log_level (str): Log level to use

    """
    # Suppress httpx logging from API requests
    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.getLogger("httpcore").setLevel(logging.WARNING)
    logging.basicConfig(
        level=log_level, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    # need to silence entrywise
    logging.getLogger("hyperopt").setLevel(logging.WARNING)  # or logging.ERROR


def serialize(x: np.ndarray) -> list:
    """Serialize a numpy array to a list.

    Args:
        x (np.ndarray): The numpy array to serialize.

    Returns:
        list: The serialized numpy array.

    """
    if np.any(np.isnan(x)):
        return []
    else:
        return x.tolist()  # type: ignore



"""Dataset loader for the HeartSteps V1 dataset.

Source: https://github.com/klasnja/HeartStepsV1
License: CC-BY-4.0
Paper Reference:
    Klasnja, P., Smith, S., Seewald, N. J., Lee, A., Hall, K., Luers, B., Hekler, E. B., & Murphy, S. A. (2019).
    Efficacy of Contextually Tailored Suggestions for Physical Activity: A Micro-randomized Optimization Trial of HeartSteps.
    Annals of Behavioral Medicine, 53(6), 573–582.
    https://doi.org/10.1093/abm/kay067
"""

from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.datasets.dataloader_factory import register_dataset
from datetime import timedelta
import numpy as np
import pandas as pd
from typing import Any
import warnings
import logging
from joblib import Memory
from nsquared.data_types import DataType


memory = Memory(".joblib_cache", verbose=2)

logger = logging.getLogger(__name__)

params = {
    "freq": (str, "5min", "Frequency of step count samples"),
    "participants": (int, 37, "Number of participants in the study"),
    "max_study_day": (int, 52, "Maximum number of study days to include"),
    "num_measurements": (int, 12, "Number of measurements per decision point"),
}


@register_dataset("heartsteps", params)
class HeartStepsDataLoader(NNDataLoader):
    """Data from the Heartsteps V1 study formatted into a matrix or tensor.
    To initialize with default settings, use: NNData.create("heartsteps").

    """

    urls = {
        "jbsteps.csv": "https://raw.githubusercontent.com/klasnja/HeartStepsV1/refs/heads/main/data_files/jbsteps.csv",
        "suggestions.csv": "https://raw.githubusercontent.com/klasnja/HeartStepsV1/refs/heads/main/data_files/suggestions.csv",
    }

    def __init__(
        self,
        agg: str = "mean",
        freq: str = "5min",
        participants: int = 37,
        max_study_day: int = 52,
        num_measurements: int = 12,
        log_transform: bool = True,
        **kwargs: Any,
    ):
        """Initializes the HeartSteps data loader.

        Args:
        ----
            agg: Aggregation method to use to create scalar dataset. Default: "mean".
            freq: Frequency of step count samples. Default: "5min".
            participants: Number of participants to include. Default: 37 (maximum).
            max_study_day: Maximum study day. Default: 52.
            num_measurements: Number of measurements taken after each decision point. Default: 12.
                Note: it is recommended that you use no more than 1-2 hours of data after each decision point to avoid overlap with the next decision point.
                This means that freq (in minutes) * num_measurements should be approximately 60.
            log_transform: Whether to apply log transformation to the data. Default: True.
            kwargs: Additional keyword arguments.

        """
        super().__init__(
            agg=agg,
            **kwargs,
        )
        self.freq = freq
        self.participants = participants
        self.max_study_day = max_study_day
        self.num_measurements = num_measurements
        self.data = None
        self.mask = None
        self.log_transform = log_transform

    def process_data_scalar(self, agg: str = "mean") -> tuple[np.ndarray, np.ndarray]:
        """Process the data into scalar setting. Note that this implementation is specific to HeartSteps as it calls upon functions that do specific HeartSteps data processing.

        Args:
        ----
            agg: Aggregation method to use. Default: "mean". Options: "mean", "sum", "median", "std", "variance"

        Returns:
        -------
            data: 2d processed data matrix of floats
            mask: Mask for processed data

        """
        df_steps, df_suggestions = self._load_data()
        data, _, mask = self._proc_dist_data(df_steps, df_suggestions)
        if agg == "mean":
            data = np.nanmean(data, axis=2)
        elif agg == "sum":
            data = np.nansum(data, axis=2)
        elif agg == "median":
            data = np.nanmedian(data, axis=2)
        elif agg == "std":
            data = np.nanstd(data, axis=2)
        elif agg == "variance":
            data = np.nanvar(data, axis=2)
        else:
            raise ValueError(
                "agg must be one of 'mean', 'sum', 'median', 'std', or 'variance'"
            )

        data = np.squeeze(data)
        self.data = data
        self.mask = mask
        return data, mask

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Process the data into distribution setting. Note that this implementation is specific to HeartSteps as it calls upon functions that do specific HeartSteps data processing.

        Args:
            data_type: Data type to process. Default: None.

        Returns:
            data: 4d processed data tensor of floats
            mask: Mask for processed data

        """
        df_steps, df_suggestions = self._load_data()

        _, data2d, mask = self._proc_dist_data(df_steps, df_suggestions)
        self.data = data2d
        self.mask = mask
        return data2d, mask

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state as a dictionary. For HeartSteps, this includes the data, masking matrix, and the custom parameters (if include_metadata == True

        If the data and mask are None, then the data has not been processed yet. Call process_data_scalar() or process_data_distribution() to process the data first.

        Args:
            include_metadata (bool): Whether to include metadata in the dictionary. Default: False. The metadata for HeartSteps is currently empty.

        """
        full_state = {
            "data": self.data,
            "mask": self.mask,
            "custom_params": {
                "freq": self.freq,
                "participants": self.participants,
                "max_study_day": self.max_study_day,
                "num_measurements": self.num_measurements,
            },
        }
        return full_state

    ## HELPER FUNCTIONS
    @classmethod
    @memory.cache
    def _load_data(cls) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Load the data from the remote source through urls

        Returns:
            df_steps: pd.DataFrame
            df_suggestions: pd.DataFrame

        """
        logger.info("Retrieving data from url...")
        jp_path = cls.urls["jbsteps.csv"]
        sug_path = cls.urls["suggestions.csv"]

        df_steps = pd.read_csv(jp_path, low_memory=False)
        df_suggestions = pd.read_csv(sug_path, low_memory=False)
        return df_steps, df_suggestions

    @staticmethod
    @memory.cache
    def _transform_dnn(
        df: Any,
        users: int = 37,
        max_study_day: int = 52,
        day_dec: int = 5,
        num_measurements: int = 12,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Iteratively transform the processed HeartSteps data into a 4d tensor"""
        warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
        final_M = np.zeros((users, max_study_day, day_dec, num_measurements))
        final_A = np.zeros((users, max_study_day, day_dec))

        for user in range(1, users + 1):
            for day in range(1, max_study_day + 1):
                for slot in range(1, day_dec + 1):
                    try:
                        df_uds = df.loc[pd.IndexSlice[(user, day, slot)]]
                        ind = df.index.get_indexer_for(df_uds.index)[0]
                        df_rng = df.iloc[
                            np.arange(ind, min(ind + num_measurements, len(df)))
                        ]
                        if df.iloc[ind]["avail"]:
                            # only take send.sedentary as the treatment indicator, could use send.active later
                            val = df.iloc[ind]["send.sedentary"]
                            conv_val = val if ~np.isnan(val) else df.iloc[ind]["send"]
                            final_A[user - 1, day - 1, slot - 1] = int(conv_val)
                        else:
                            final_A[user - 1, day - 1, slot - 1] = 2

                        measurements = df_rng["steps"].to_numpy()
                        if len(measurements) == num_measurements:
                            final_M[user - 1, day - 1, slot - 1] = measurements
                        else:
                            m_pad = np.pad(
                                measurements,
                                (0, num_measurements - len(measurements)),
                                constant_values=np.nan,
                            )
                            final_M[user - 1, day - 1, slot - 1] = m_pad
                    except KeyError:
                        final_A[user - 1, day - 1, slot - 1] = 0
                        final_M[user - 1, day - 1, slot - 1] = np.full(
                            num_measurements, np.nan
                        )
        final_M = final_M.reshape((users, max_study_day * day_dec, num_measurements))
        final_A = final_A.reshape((users, max_study_day * day_dec))
        return final_M, final_A

    @staticmethod
    def _get_mode(x: pd.Series) -> Any:
        if len(pd.Series.mode(x) > 1):
            return pd.Series.mode(x, dropna=False)[0]
        else:
            return pd.Series.mode(x, dropna=False)

    def _reind_id(self, df_u: Any) -> pd.DataFrame:
        """Function to reindex the data to include all time points for each user."""
        d_rnge = pd.date_range(
            min(df_u.index.astype("datetime64[ns]")),
            max(df_u.index.astype("datetime64[ns]")) + timedelta(days=1),
            normalize=True,
            inclusive="both",
            freq=self.freq,
        )
        d_rnge = d_rnge[d_rnge.indexer_between_time("00:00", "23:55")]
        # print(rng)
        df_reind = df_u.reindex(d_rnge)
        df_reind["user.index"] = df_reind["user.index"].ffill().bfill()
        df_reind["study.day.nogap"] = df_reind["study.day.nogap"].bfill().ffill()
        df_reind["steps"] = df_reind["steps"].fillna(0)
        return df_reind

    @staticmethod
    def _take_range(df: Any, range: int) -> pd.DataFrame:
        idx = df.index.get_indexer_for(df[pd.notna(df["sugg.select.slot"])].index)
        ranges = [np.arange(i, min(i + range + 1, len(df))) for i in idx]
        return df.iloc[np.concatenate(ranges)]

    @staticmethod
    def _create_slots(df: Any) -> pd.DataFrame:
        most_rec_slot = 0.0
        for ind, row in df.iterrows():
            curr_slot = row["sugg.select.slot"]
            if not np.isnan(curr_slot):
                if (
                    most_rec_slot != 5.0
                    and curr_slot != most_rec_slot + 1
                    and most_rec_slot != 0.0
                ):
                    df.at[ind, "new_slot"] = most_rec_slot + 1
                    most_rec_slot += 1
                elif most_rec_slot == 5.0 and curr_slot != 1:
                    df.at[ind, "new_slot"] = 1
                    most_rec_slot = 1
                else:
                    df.at[ind, "new_slot"] = curr_slot
                    most_rec_slot = curr_slot
        return df

    @staticmethod
    def _study_day(df: Any) -> pd.DataFrame:
        most_rec_slot = 1.0
        curr_study_day = 1
        for ind, row in df.iterrows():
            curr_slot = row["new_slot"]
            if not np.isnan(curr_slot):
                if most_rec_slot == 5.0 and curr_slot == 1:
                    curr_study_day += 1
                most_rec_slot = curr_slot
                df.at[ind, "study_day"] = curr_study_day
        return df

    @staticmethod
    @memory.cache
    def _group_steps(df: pd.DataFrame, freq: str) -> pd.DataFrame:
        return df.groupby(
            [
                # TODO: (Caleb) resolve this Grouper pyright error - says no parameter named 'label' but pd.Grouper param list has 'label'.
                # Main issue is that TimeGrouper (which has the param label) was deprecated but is still used under the hood
                # so the param label is not explicitly exposed in the Grouper init definition but is still accepted/used.
                pd.Grouper(freq=freq, level="steps.utime", label="right"),  # pyright: ignore
                pd.Grouper(level="user.index"),
            ],
            sort=False,
        ).agg(
            {
                "steps": "sum",
                "study.day.nogap": lambda x: HeartStepsDataLoader._get_mode(x),
            }
        )

    def _proc_dist_data(
        self, df_steps: pd.DataFrame, df_suggestions: pd.DataFrame
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Process the data into distribution setting."""
        logger.debug("Processing HeartSteps V1...")
        # get relevant cols and reformatting
        df_steps = pd.DataFrame(
            df_steps[["user.index", "steps.utime", "steps", "study.day.nogap"]]
        )
        df_steps = df_steps.copy()
        df_steps["steps.utime"] = pd.to_datetime(df_steps["steps.utime"])
        # create multi-index
        df_steps = df_steps.set_index(["user.index", "steps.utime"])

        # get relevant cols and reformatting
        df_sugg_sel = df_suggestions[
            [
                "user.index",
                "decision.index.nogap",
                "sugg.select.utime",
                "sugg.decision.utime",
                "sugg.select.slot",
                "avail",
                "send",
                "send.active",
                "send.sedentary",
            ]
        ]
        df_sugg_sel = df_sugg_sel.copy()
        df_sugg_sel["sugg.decision.utime"] = pd.to_datetime(
            df_sugg_sel["sugg.decision.utime"]
        )
        # TODO: (Caleb) Resolve pyright error with dropna function
        df_sugg_sel = df_sugg_sel.dropna(  # pyright: ignore
            subset=["sugg.decision.utime", "sugg.select.utime", "user.index"]
        )

        # group the step data by five minute intervals
        df_freq = HeartStepsDataLoader._group_steps(df_steps, self.freq).reset_index()

        df_freq_ind = df_freq.set_index("steps.utime")

        # expand the step data to include all time points
        # df_expand5min = df_5min_ind.groupby("user.index", group_keys=False).apply(
        #     lambda df_u: _reind_id(df_u)
        # )

        result_dfs = []
        user_indices = df_freq_ind["user.index"].unique()
        # process each user
        for user_idx in user_indices:
            # filter for just this user
            df_u = df_freq_ind[df_freq_ind["user.index"] == user_idx].copy()
            reindexed_df = self._reind_id(df_u)
            result_dfs.append(reindexed_df)
        df_expandfreq = pd.concat(result_dfs)

        df_expandfreq = df_expandfreq.reset_index(names="steps.utime")
        df_expandfreq["user.index"] = df_expandfreq["user.index"].astype("int64")

        # merge the step data with the notification data
        df_merged = (
            pd.merge_asof(
                df_expandfreq.sort_values(by="steps.utime"),
                df_sugg_sel.sort_values(by="sugg.decision.utime"),
                left_on="steps.utime",
                right_on="sugg.decision.utime",
                by="user.index",
                # TODO: (Caleb) Resolve pyright error with pd.Timedelta. This is due to another incompatibility in the pandas type specification.
                # tolerance does not accept NaT, but Timedelta could return NaT. Pandas documentation uses pd.Timedelta in this way exactly, so unsure of solution.
                tolerance=pd.Timedelta(self.freq),  # pyright: ignore
                allow_exact_matches=False,
                direction="backward",
            )
            .sort_values(by=["user.index", "steps.utime"])
            .reset_index(drop=True)
        )
        df_merged["sugg.select.slot"] = np.where(
            df_merged["decision.index.nogap"].isna(),
            np.nan,
            df_merged["sugg.select.slot"],
        )

        # get num_measurement rows after each notification period (1 hour of observations by default args)
        unique_users = df_merged["user.index"].unique()
        result_dfs = []

        for user_idx in unique_users:
            df_user = df_merged[df_merged["user.index"] == user_idx]
            df_user_range = HeartStepsDataLoader._take_range(
                df_user, self.num_measurements
            )
            result_dfs.append(df_user_range)

        df_merged_cut = pd.concat(result_dfs).reset_index(drop=True)
        df_merged_cut_nd = df_merged_cut.drop_duplicates()

        # set up column for study day
        df_merged_cut_nd = df_merged_cut_nd.copy()
        df_merged_cut_nd["study_day"] = np.nan
        df_merged_cut_nd["new_slot"] = np.nan

        # align decision points
        unique_users = df_merged_cut_nd["user.index"].unique()
        slot_results = []

        for user_idx in unique_users:
            df_user = df_merged_cut_nd[df_merged_cut_nd["user.index"] == user_idx]
            user_slots = HeartStepsDataLoader._create_slots(df_user)
            slot_results.append(user_slots)

        df_slot = pd.concat(slot_results)

        # Second groupby operation: study_day
        unique_users_slot = df_slot["user.index"].unique()
        study_day_results = []

        for user_idx in unique_users_slot:
            df_user = df_slot[df_slot["user.index"] == user_idx]
            user_study_day = HeartStepsDataLoader._study_day(df_user)
            study_day_results.append(user_study_day)

        df_study_day = pd.concat(study_day_results)

        # create unique index
        df_final = df_study_day.set_index(["user.index", "study_day", "new_slot"])

        # transform into 4d tensor + mask
        data, mask = HeartStepsDataLoader._transform_dnn(
            df_final,
            users=self.participants,
            max_study_day=self.max_study_day,
            num_measurements=self.num_measurements,
        )
        N, T = mask.shape
        data2d = np.empty([N, T], dtype=object)
        if self.log_transform:
            data = np.log(data + 1)

        # to align with 4d structure
        data = data[:, :, :, np.newaxis]

        for i in range(N):
            for j in range(T):
                data2d[i, j] = data[i, j].flatten()
        return data, data2d, mask



"""Abstract class for experiment/real data loading. There are two key attributes:"""

from abc import ABC, abstractmethod
import numpy as np
from typing import Any
from nsquared.data_types import DataType


class NNDataLoader(ABC):
    # this is an abstract attribute to contain the URLs of the data of subclassers
    urls: dict
    supported_aggs = ["mean", "sum", "median", "std", "variance"]

    def __init__(
        self,
        agg: str = "mean",
        save_processed: bool = False,
        **kwargs: Any,
    ):
        """Initializes the data loader.

        Args:
            agg (str): aggregation method to use to create scalar dataset. Default: "mean".
            save_processed (bool): whether to save the processed data.  Default: False.
            **kwargs: additional arguments to be passed to the subclass.

        """
        self.save_processed = save_processed
        if agg not in self.supported_aggs:
            raise ValueError(
                f"Aggregation method {agg} not supported. Supported methods: {self.supported_aggs}"
            )
        self.agg = agg

    @abstractmethod
    def process_data_scalar(self) -> tuple[np.ndarray, np.ndarray]:
        """Processes the data into scalar matrix setting. Returns a tuple of ndarrays representing 2d data matrix and mask.

        Returns:
            data: 2d data matrix
            mask: 2d mask matrix

        """
        raise NotImplementedError

    @abstractmethod
    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Processes the data into distributional setting. Returns a tuple of ndarrays representing 4d distributional data and mask.

        Returns:
            data: 4d distributional data matrix
            mask: 2d mask matrix

        """
        raise NotImplementedError

    @abstractmethod
    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """A helpful debugging tool: returns the full state of the data loader as a dictionary.

        Args:
            include_metadata (bool): whether to include metadata in the dictionary. Default: False.

        """
        raise NotImplementedError



"""Dataset loader for the PromptEval (MMLU) dataset.

Source1: https://huggingface.co/datasets/PromptEval
Source2: https://github.com/kyuseongchoi5/EfficientEval_BayesOpt
Paper Reference for transformation implementation:
    Felipe Maia Polo et al.
    "Efficient multi-prompt evaluation of LLMs."
    Neurips, 2024.
    https://arxiv.org/pdf/2405.17202
"""

from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.datasets.dataloader_factory import register_dataset
from nsquared.data_types import (
    DataType,
    DistributionKernelMMD,
    DistributionWassersteinSamples,
    DistributionWassersteinQuantile,
)
import numpy as np
import pandas as pd
from typing import Any, cast
import logging
from joblib import Memory
from datasets import load_dataset, Dataset
from tqdm import tqdm


memory = Memory(".joblib_cache", verbose=0)
logger = logging.getLogger(__name__)

params = {
    "tasks": (
        list[str],
        None,
        "List of tasks to evaluate on. By default, returns all.",
    ),
    "models": (
        list[str],
        None,
        "List of models to evaluate by. By default, returns all.",
    ),
    "seed": (int, None, "Random seed for reproducibility"),
    "propensity": (float, None, "Proportion of data to keep"),
}


@register_dataset("prompteval", params)
class PromptEvalDataLoader(NNDataLoader):
    """Data from the PromptEval study formatted into a matrix or tensor.
    To initialize with default settings, use: NNData.create("prompteval").

    """

    MODELS = [
        "meta_llama_llama_3_8b",
        "meta_llama_llama_3_8b_instruct",
        "meta_llama_llama_3_70b_instruct",
        "codellama_codellama_34b_instruct",
        "google_flan_t5_xl",
        "google_flan_t5_xxl",
        "google_flan_ul2",
        "ibm_mistralai_merlinite_7b",
        "mistralai_mixtral_8x7b_instruct_v01",
        "mistralai_mistral_7b_instruct_v0_2",
        "google_gemma_7b",
        "google_gemma_7b_it",
        "tiiuae_falcon_40b",
        "mistralai_mistral_7b_v0_1",
        "tiiuae_falcon_180b",
    ]

    TASKS = [
        "college_mathematics",
        "miscellaneous",
        "moral_disputes",
        "jurisprudence",
        "moral_scenarios",
        "college_chemistry",
        "professional_medicine",
        "clinical_knowledge",
        "abstract_algebra",
        "nutrition",
        "professional_psychology",
        "high_school_government_and_politics",
        "high_school_us_history",
        "high_school_chemistry",
        "high_school_macroeconomics",
        "management",
        "conceptual_physics",
        "philosophy",
        "electrical_engineering",
        "high_school_psychology",
        "medical_genetics",
        "high_school_geography",
        "high_school_statistics",
        "international_law",
        "elementary_mathematics",
        "high_school_physics",
        "world_religions",
        "high_school_european_history",
        "formal_logic",
        "security_studies",
        "sociology",
        "high_school_biology",
        "us_foreign_policy",
        "high_school_microeconomics",
        "college_medicine",
        "college_computer_science",
        "logical_fallacies",
        "high_school_computer_science",
        "anatomy",
        "econometrics",
        "astronomy",
        "college_biology",
        "virology",
        "professional_accounting",
        "college_physics",
        "high_school_world_history",
        "business_ethics",
        "global_facts",
        "public_relations",
        "marketing",
        "human_aging",
        "professional_law",
        "high_school_mathematics",
        "prehistory",
        "machine_learning",
        "computer_security",
        "human_sexuality",
    ]

    def __init__(
        self,
        tasks: list[str] | None = None,
        models: list[str] | None = None,
        seed: int | None = None,
        propensity: float = 1.0,  # Default to 1.0 (keeping all data)
        n_examples_per_task: int = 100,
        **kwargs: Any,
    ):
        """Initializes the PromptEval data loader.

        Args:
        ----
            tasks: benchmark tasks to evaluate on. Default: None (use all tasks).
            models: models that are evaluated on for each tasks. Default: None (use all models).
            seed: Random seed for reproducibility. Default: None
            propensity: Proportion of data to keep. Default: 1.0
            n_examples_per_task: Number of examples to sample (with replacement) from each task
                in the distributional setting. Default: 100.
            kwargs: Additional keyword arguments.

        """
        super().__init__(
            **kwargs,
        )
        self.tasks = self.TASKS if tasks is None else tasks
        self.models = self.MODELS if models is None else models
        self.propensity = propensity
        self.seed = seed
        if self.seed is not None:
            np.random.seed(
                seed=self.seed
            )  # instantiate random seed if provided but do it only once here
        self.n_examples_per_task = n_examples_per_task

        # print('Loading dataset')
        # for task in tqdm(self.tasks):
        #     # Download the dataset to cache
        #     load_dataset("PromptEval/PromptEval_MMLU_correctness", name=task)

    @staticmethod
    @memory.cache
    def load_config_data(
        task: str, model: str, n_examples_per_task: int = 100, seed: int | None = None
    ) -> pd.DataFrame | None:
        """Load and process data for a specific configuration.

        Args:
            task: Name of the task to load.
            model: Name of the model to load.
            n_examples_per_task: Number of examples to sample (with replacement) from each task. Default: 100.
            seed: Random seed for reproducibility. Default: None.

        Returns:
            DataFrame containing the processed data or None if loading fails.

        """
        # Load the dataset with the specific config
        # NOTE: dataset is a Dataset
        # NOTE: HF automatically downloads the dataset if it's not already downloaded
        # and caches it in ~/.cache/huggingface/hub/
        dataset = cast(
            Dataset,
            load_dataset(
                "PromptEval/PromptEval_MMLU_correctness",
                name=task,
                split=model,
                num_proc=12,
                keep_in_memory=True,
            ),
        )
        # rows are format templates, columns are examples
        df_table: pd.DataFrame = cast(pd.DataFrame, dataset.to_pandas())
        # sample columns with replacement
        df_table = (
            df_table.transpose()
            .sample(n=n_examples_per_task, replace=True, random_state=seed)
            .transpose()
            .reset_index()
        )
        # melt the table into rows for every format-example pair
        df_long = pd.melt(
            df_table,
            id_vars=["index"],
            var_name="example",
            value_name="correctness",
        )
        df_long["model"] = model
        df_long["task"] = task
        df_long = df_long.rename(columns={"index": "format"})
        return df_long

    def process_data_scalar(self) -> tuple[np.ndarray, np.ndarray]:
        """Processes the data into scalar setting. This implementation is applicable when generating (template * example) matrix, while fixing model and task.

        Returns
        -------
            data: 2d processed data matrix of floats (in this case, each entry is boolean as the metric is correctness)
            mask: Mask for processed data

        """
        if not self.tasks or not self.models:
            raise ValueError("Tasks and models must be specified")

        assert len(self.models) == 1, "Only one model is supported in scalar mode"
        assert len(self.tasks) == 1, "Only one task is supported in scalar mode"

        model = self.models[0]
        task = self.tasks[0]
        propensity = self.propensity

        ds = cast(
            Dataset,
            load_dataset(
                "PromptEval/PromptEval_MMLU_correctness",
                name=task,
                split=model,
                keep_in_memory=True,
                num_proc=8,
            ),
        )
        df = cast(pd.DataFrame, ds.to_pandas())
        mask = np.random.binomial(1, propensity, size=df.shape)

        data = df.to_numpy(dtype=float)
        # data[mask == 0] = np.nan
        self.data = data
        self.mask = mask
        return data, mask

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Process the data into distributional setting.

        Args:
        ----
            data_type: Data type to use for the data. Default: None.

        Returns:
        -------
            data: numpy array of shape (n_tasks, n_models, n_templates, 1)
                if data_type is None or DistributionKernelMMD.
                numpy array of shape (n_tasks, n_models, n_templates)
                if data_type is DistributionWassersteinSamples or DistributionWassersteinQuantile.
                Entry is the average correctness across examples for a given model-task-template combination
            mask: numpy array of shape (n_tasks, n_models, n_templates, 1)
                Entry is 1 if the example is kept, 0 otherwise

        """
        if not self.tasks or not self.models:
            raise ValueError("Tasks and models must be specified")

        models = self.models
        tasks = self.tasks
        propensity = self.propensity

        # Load the data for the multiple config and model
        df_list = []
        for task in tqdm(tasks):
            for model in models:
                df = self.load_config_data(
                    task, model, self.n_examples_per_task, self.seed
                )
                df_list.append(df)

        df = pd.concat(df_list, ignore_index=True)

        # compute the mean correctness across examples for each model-task-format combination
        df = df.pivot_table(
            index=["model", "task", "format"], values="correctness", aggfunc="mean"
        )
        # aggregate the correctness values into a list for each model-task combination
        df = df.groupby(["model", "task"]).agg(list)
        # unstack the task and model index into separate columns
        df = df.unstack()
        # drop the multi-index correctness columns
        df = df.droplevel(0, axis=1)
        # convert lists to numpy arrays of type float
        if data_type is None or isinstance(data_type, DistributionKernelMMD):
            entry_shape = (-1, 1)
        elif isinstance(data_type, DistributionWassersteinSamples) or isinstance(
            data_type, DistributionWassersteinQuantile
        ):
            entry_shape = (-1,)
        else:
            raise ValueError(f"Invalid data type: {data_type}")

        data = np.array(
            [np.array(x, dtype=float) for x in df.values.flatten()]  # type: ignore
        ).reshape(df.shape + entry_shape)

        # Simulate MCAR missingness
        mask = np.random.binomial(1, propensity, size=data.shape[:2])
        # TODO: change this back later
        # data[mask == 0] = np.nan

        self.data = data
        self.mask = mask
        return data, mask

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state as a dictionary.

        Args:
            include_metadata (bool): Whether to include metadata in the dictionary. Default: False.

        Returns:
            dict: Dictionary containing the data and mask

        """
        full_state = {
            "data": self.data,
            "mask": self.mask,
        }
        return full_state



import numpy as np


def usvt(A: np.ndarray, eta: float = 0.0001) -> np.ndarray:
    """Compute the USVT matrix imputations

    Args:
    ----
    A : a N x T matrix with missing values denoted as nan
    eta : a small positive (or 0) that affects the singular value threshold.
            by default 0

    Returns:
    -------
    A_hat : the USVT estimate of A

    """
    # later should generalize for all matrices (i.e. transpose if m > n)

    N, T = A.shape

    # data prep
    # of observed values,
    a = np.nanmin(A)
    b = np.nanmax(A)

    A_scaled = (A - ((a + b) / 2)) / ((b - a) / 2)
    Y_scaled = np.where(np.isnan(A_scaled), 0, A_scaled)

    p_hat = 1 - (np.count_nonzero(np.isnan(A)) / np.size(A))

    # currently follow Chatterjee 2015 threshold value
    threshold = (2 + eta) * np.sqrt(T * p_hat)

    u, s, vt = np.linalg.svd(Y_scaled, full_matrices=False)

    s_mask = s >= threshold
    W = (u[:, s_mask] @ np.diag(s[s_mask]) @ vt[s_mask, :]) / p_hat

    # cap to [-1, 1]
    W = W.clip(-1, 1)
    # rescale:
    res = (W * ((b - a) / 2)) + (a + b) / 2

    return res



"""Kernel function implementations

TODO: add linear, square, and exponential kernels from https://github.com/calebchin/DistributionalNearestNeighbors/blob/main/src/kernel_nn.py
"""

import numpy as np


def laplace(dists: np.ndarray, eta: float) -> np.ndarray:
    """Compute the Laplace kernel given a distance matrix.

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix.

    """
    assert eta > 0
    dists = np.where(dists < 0, 0, dists)
    gamma = 1.0 / eta
    dists *= -gamma
    kernel_mat = np.exp(dists)
    return kernel_mat


def gaussian(dists: np.ndarray, eta: float) -> np.ndarray:
    """Compute the Gaussian kernel given a distance matrix.

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix.

    """
    assert eta > 0
    dists = np.where(dists < 0, 0, dists)
    gamma = 0.5 / eta**2
    dists *= -gamma
    kernel_mat = np.exp(dists)
    return kernel_mat


def singular(dists: np.ndarray, eta: float) -> np.ndarray:
    """Compute the Singular kernel given a distance matrix.

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix.

    """
    dists /= eta
    a = 0.49
    # replace divide by zero with nan
    dists = np.where(dists == 0, np.nan, dists)
    kernel_matrix = np.power(dists, -a) * np.maximum(0, 1 - dists) ** 2
    return kernel_matrix


def singular_box(dists: np.ndarray, eta: float) -> np.ndarray:
    """Compute the Singular Box kernel given a distance matrix.

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix.

    """
    assert eta > 0
    dists /= eta
    a = 0.49
    dists = np.where(dists == 0, np.nan, dists)
    kernel_mat = np.power(dists, -a) * np.where(dists <= 1, 1, 0)
    return kernel_mat


def box(dists: np.ndarray, eta: float) -> np.ndarray:
    """Compute the Box kernel given a distance matrix.

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix with values 1 where the distance is within the bandwidth, otherwise 0.

    """
    assert eta > 0
    dists /= eta
    kernel_mat = np.where(dists <= 1, 1, 0)
    return kernel_mat


def epanechnikov(dists: np.ndarray, eta: float) -> np.ndarray:
    r"""Compute the Epanechnikov kernel given a distance matrix.
        Epanechnikov kernel: \kappa(u) = 3/4 * (1 - u^2) for u in [-1, 1],
        where u = dists / eta

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix with values computed using the Epanechnikov kernel function.

    """
    assert eta > 0
    dists /= eta
    kernel_mat = np.where(dists <= 1, 0.75 * (1 - dists**2), 0)
    return kernel_mat


def wendland(dists: np.ndarray, eta: float) -> np.ndarray:
    r"""Compute the Wendland kernel given a distance matrix.
        Wendland kernel: \kappa(u) = (1 - u)_+ for u in [0, 1],
        where u = dists / eta

    Args:
        dists (np.ndarray): Distance matrix.
        eta (float): Bandwidth parameter for the kernel.

    Returns:
        np.ndarray: Kernel matrix with values computed using the Wendland kernel function.

    """
    assert eta > 0
    dists /= eta
    kernel_mat = np.where(dists <= 1, 1 - dists, 0)
    return kernel_mat
